"""
    RemoteAgentAdapter - Adapter to access the Remote Services
"""
import io
import logging
from urllib.parse import urljoin

import requests

from starfish.utils.crypto_hash import hash_sha3_256

logger = logging.getLogger(__name__)


class ResponseWrapper():
    """
    Response object returned by different test and production
    systems are not the same, and have different properties to obtain
    data and JSON data.

    This class tries to obtain the correct call to get the data from the
    Response object.

    """
    def __init__(self, response):
        self._response = response

    @property
    def json(self):
        if hasattr(self._response, 'get_json'):
            data = self._response.get_json()
        elif hasattr(self._response, 'json'):
            data = self._response.json()
        else:
            TypeError('cannot get the json property from response object')
        return data

    @property
    def data(self):
        if hasattr(self._response, 'content'):
            data = self._response.content
        elif hasattr(self._response, 'data'):
            data = self._response.data
        else:
            raise TypeError('Cannot find correct response data')
        return data


class RemoteAgentAdapter():

    def __init__(self, http_client=None):
        self._http_client = http_client
        if self._http_client is None:
            self._http_client = requests

    def register_asset(self, metadata, url, authorization_token=None):
        """
        Register an asset with the agent storage server
        :param metadata: metadata to write to the storage server

        :return: A dict of the following items
        [0] asset_id.
        [1] did of the asset.
        [2] metadata in text format, that was saved.

        :type: dict
        """
        result = None
        saved_asset_id = self.save_metadata(metadata, url, authorization_token)
        result = {
                'asset_id': saved_asset_id,
                'metadata': metadata,
                'hash': hash_sha3_256(metadata)
            }
        return result

    def get_metadata_list(self, url, authorization_token=None):
        """ return a list of metadata stored on this agent """
        url = urljoin(f'{url}/', 'index')
        logger.debug(f'metadata list url {url}')
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'metadata asset response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)

    def save_metadata(self, metadata, url, authorization_token=None):
        """save metadata to the agent server, using the asset_id and metadata"""
        url = urljoin(f'{url}/', 'data')
        logger.debug(f'metadata save url {url}')
        headers = RemoteAgentAdapter.create_headers('text/plain', authorization_token)
        response = self._http_client.post(url, data=metadata, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'metadata asset response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def create_listing(self, listing_data, asset_id, url, authorization_token=None):
        data = {
            'assetid': asset_id,
            'info': listing_data,
        }

        url = urljoin(f'{url}/', 'listings')
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.post(url, json=data, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            logger.debug('listing response returned: ' + str(data))
            return data
        else:
            msg = f'listing response failed: {response.status_code} {response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def download_asset(self, asset_id, url, authorization_token=None):
        url = urljoin(f'{url}/', asset_id)

        headers = RemoteAgentAdapter.create_headers('application/octet-stream', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).data
            return data
        else:
            msg = f'GET assets response failed: {response.status_code} {response}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_listing(self, listing_id, url, authorization_token=None):
        url = urljoin(f'{url}/', f'listings/{listing_id}')
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_listings(self, url,  authorization_token=None):
        url = urljoin(f'{url}/', 'listings')
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            return ResponseWrapper(response).json
        else:
            msg = f'GET listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def update_listing(self, listing_id, data, url, authorization_token=None):
        url = urljoin(f'{url}/', f'listings/{listing_id}')

        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.put(url, json=data, headers=headers)
        if response and response.status_code == requests.codes.ok:
            return True
        else:
            msg = f'PUT listings response failed: {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def upload_asset_data(self, asset_id, data, url, authorization_token=None):
        url = urljoin(f'{url}/', asset_id)

        logger.debug(f'uploading data to {url}')
        files = {
            'file': (asset_id, io.BytesIO(data), 'application/octet-stream'),
        }
        headers = RemoteAgentAdapter.create_headers(None, authorization_token)
        response = self._http_client.post(url, files=files, headers=headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == requests.codes.created):
            return True
        else:
            msg = f'upload asset response failed: {response.status_code}:{response.text}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def read_metadata(self, asset_id, url, authorization_token=None):
        """read the metadata from a service agent using the asset_id"""

        result = None
        url = urljoin(f'{url}/', f'data/{asset_id}')
        logger.debug(f'metadata read url {url}')
        headers = RemoteAgentAdapter.create_headers('text/plain', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).data
            result = {
                'asset_id': asset_id,
                'metadata_text': data,
            }
            result['hash'] = hash_sha3_256(result['metadata_text'])
            # convert to str if bytes
            if isinstance(result['metadata_text'], bytes):
                result['metadata_text'] = data.decode('utf-8')
        else:
            logger.warning(f'metadata asset read {asset_id} response returned {response} for {url}')
        return result

    def purchase_asset(self, purchase, url, authorization_token=None):
        """record purchase"""
        url = urljoin(f'{url}/', 'purchases')
        logger.debug(f'market url for purchases {url}')
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.post(url, json=purchase, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            logger.debug(f'purchase response returned {data}')
            return data
        else:
            msg = f'purchase response failed: {response.status_code}'
            logger.error(msg)
            # NOTE the agent may return error information in additon to a 500
            if response:
                if hasattr(response, 'get_json'):
                    data = response.get_json()
                elif hasattr(response, 'json'):
                    data = response.json()
                else:
                    TypeError('cannot get the json property from response object')
                logger.error(f'purchase response returned {data}')
            raise ValueError(msg)
        return None

    def invoke(self, asset_id, inputs, url, authorization_token=None):
        """

        call the invoke based on the asset_id without leading 0x
        """

        url = urljoin(f'{url}/', asset_id)
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.post(url, json=inputs, headers=headers)
        if response and (response.status_code == requests.codes.ok or response.status_code == 201):
            data = ResponseWrapper(response).json
            logger.debug('invoke response returned: ' + str(data))
            return data
        else:
            msg = f'invoke response failed: {response.status_code} {response} for {url}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_job(self, job_id, url, authorization_token=None):
        url = urljoin(f'{url}/', job_id)
        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response and response.status_code == requests.codes.ok:
            data = ResponseWrapper(response).json
            return data
        else:
            msg = f'GET job response failed: {response.status_code} for {url}'
            logger.error(msg)
            raise ValueError(msg)
        return None

    def get_authorization_token(self, username, password, url):
        """Get an agent authorization token (create one if needed).
        Throws exception on error."""
        response = self._http_client.get(url, auth=(username, password))
        token = None
        if response.status_code == 200:
            tokens = ResponseWrapper(response).json
            if len(tokens) > 0:
                token = tokens[-1]
            else:   # need to create a token
                response = self._http_client.post(url, auth=(username, password))
                if response.status_code == 200:
                    token = ResponseWrapper(response).json
                else:
                    msg = f'unable to create token, status {response.status_code}'
                    logger.error(msg)
                    raise ValueError(msg)
        else:
            msg = f'unable to get tokens, status {response.status_code}'
            logger.error(msg)
            raise ValueError(msg)
        logger.debug(f'using agent token {token}')
        return token

    def get_ddo(self, url, authorization_token=None):
        ddo_text = None
        url = urljoin(f'{url}/', '/api/ddo')
        logger.debug(f'get_ddo url {url}')

        headers = RemoteAgentAdapter.create_headers('application/json', authorization_token)
        response = self._http_client.get(url, headers=headers)
        if response.status_code == 200:
            ddo_text = ResponseWrapper(response).data.decode('utf-8')
        return ddo_text

    @property
    def http_client(self):
        return self._http_client

    @http_client.setter
    def http_client(self, value):
        """Set the http client to something other than the default `requests`"""
        self._http_client = value

    @staticmethod
    def is_metadata_valid(asset_id, metadata_text):
        """
        validate metadata, by calcualating the hash (asset_id) and compare this to the
        given asset_id, if both are equal then the metadata is valid
        :param asset_id: asset id to compare with
        :param metadata: dict of metadata to calculate the hash ( asset_id)
        :return bool True if metadata is valid for the asset_id provided
        """
        if metadata_text:
            # the calc asset_id from the metadata should be same as this asset_id
            metadata_id = RemoteAgentAdapter.get_asset_id_from_metadata(metadata_text)
            if metadata_id != asset_id:
                logger.debug(f'metdata does not match {metadata_id} != {asset_id}')
            return metadata_id == asset_id
        return False

    @staticmethod
    def get_asset_id_from_metadata(metadata_text):
        """
        return the asset_id calculated from the metadata
        :param metadata: dict of metadata to hash
        a 64 char hex string, which is the asset id
        :return 64 char hex string, with no leading '0x'
        """
        return hash_sha3_256(metadata_text)

    @staticmethod
    def create_headers(content_type=None, authorization_token=None):
        headers = {}
        if content_type:
            headers['content-type'] = content_type
        if authorization_token:
            headers['Authorization'] = f'token {authorization_token}'
        if headers == {}:
            return None
        return headers



"""

    Command 'Asset Store'

"""

import datetime
import os
import re
from typing import Any

from starfish.agent import RemoteAgent
from starfish.asset import DataAsset
from .command_base import CommandBase


DEFAULT_ASSET_NAME = 'starfish_asset'


class AssetStoreCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('store', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool to store an asset',
            help='Tool to store an asset',
        )

        parser.add_argument(
            '-u',
            '--username',
            help='Optional username to access the agent'
        )

        parser.add_argument(
            '-p',
            '--password',
            help='Optional password to access the agent'
        )

        parser.add_argument(
            '-n',
            '--name',
            default=DEFAULT_ASSET_NAME,
            help='Asset name. Default: {DEFAULT_ASSET_NAME}'
        )

        parser.add_argument(
            '--description',
            help='Asset metadata description'
        )

        parser.add_argument(
            '--date-created',
            help='Asset metadata dateCreated'
        )

        parser.add_argument(
            '--created-now',
            action='store_true',
            help='Asset metadata set the dateCreated field to the current local time'
        )

        parser.add_argument(
            '--author',
            help='Asset metadata author'
        )

        parser.add_argument(
            '--license',
            help='Asset metadata license'
        )

        parser.add_argument(
            '--copyright-holder',
            help='Asset metadata copyrightHolder'
        )

        parser.add_argument(
            '--in-language',
            help='Asset metadata inLanguage'
        )

        parser.add_argument(
            '--tags',
            help='Asset metadata tags (string comma seperated list)'
        )

        parser.add_argument(
            'agent',
            help='agent url or agent did to store the asset'
        )

        parser.add_argument(
            'filename',
            help='filename to store'
        )

        return parser

    def execute(self, args: Any, output: Any) -> None:

        if not os.path.exists(args.filename):
            output.add_line(f'cannot find file {args.filename}')
            return

        network = self.get_network(args.url)

        ddo = network.resolve_agent(args.agent, username=args.username, password=args.password)
        if not ddo:
            output.add_line(f'cannot resolve asset {args.asset}')
            return

        authentication = None
        if args.username or args.password:
            authentication = {
                'username': args.username,
                'password': args.password
            }

        agent = RemoteAgent(ddo=ddo, authentication=authentication)
        metadata = {}

        if args.description:
            metadata['description'] = args.description

        if args.created_now:
            metadata['dateCreated'] = datetime.datetime.now().isoformat()

        if args.date_created:
            metadata['dateCreated'] = args.date_created

        if args.author:
            metadata['author'] = args.author

        if args.license:
            metadata['license'] = args.license

        if args.copyright_holder:
            metadata['copyrightHolder'] = args.copyright_holder

        if args.in_language:
            metadata['inLanguage'] = args.in_language

        if args.tags:
            print('tags', args.tags)
            tag_list = re.split(r'[\s\S]+', args.tags)
            metadata['tags'] = tag_list

        if not metadata.keys():
            metadata = None

        asset = DataAsset.create_from_file(args.name, args.filename, metadata=metadata)
        asset = agent.register_asset(asset)
        agent.upload_asset(asset)
        output.add_line(f'stored asset {asset.did}')
        output.set_value('asset_did', asset.did)



"""

Test provenance


"""
import json
import re
import secrets

from starfish.asset.provenance import Provenance
from starfish.utils.did import did_generate_random


def get_activity_id(result):
    assert(result['activity'])
    activity_id = list(result['activity'].keys())[0]
    return activity_id

def assert_prefix(result):
    assert(result['prefix'])
    assert(result['prefix']['xsd'])
    assert(result['prefix']['xsd'] =='http://www.w3.org/2001/XMLSchema#')
    assert(result['prefix']['prov'])
    assert(result['prefix']['prov'] == 'http://www.w3.org/ns/prov#')
    assert(result['prefix']['dep'])
    assert(result['prefix']['dep'] == 'http://dex.sg')

def assert_activity(result, activity_id, activity_type, inputs_text=None):
    assert(result['activity'])
    assert(result['activity'][activity_id])
    assert(result['activity'][activity_id]['prov:type']['$'] == f'dep:{activity_type}')
    if inputs_text:
        assert(result['activity'][activity_id]['dep:inputs'])
        assert(result['activity'][activity_id]['dep:inputs']['$'] == inputs_text)


def assert_entity(result, asset_list=None):
    assert(result['entity'])
    assert(result['entity']['dep:this'])
    assert(result['entity']['dep:this']['prov:type'])
    assert(result['entity']['dep:this']['prov:type']['$'] == 'dep:asset')
    if asset_list:
        for asset_did in asset_list:
            assert(result['entity'][asset_did])
            assert(result['entity'][asset_did]['prov:type'])
            assert(result['entity'][asset_did]['prov:type']['$'] == 'dep:asset')
        assert(len(list(result['entity'].keys())) == len(asset_list) + 1)
    else:
        # only one record in the entity the 'dep:this'
        assert(len(list(result['entity'].keys())) == 1)

def assert_was_generated_by(result, activity_id):
    assert(result['wasGeneratedBy'])
    id_dep = list(result['wasGeneratedBy'].keys())[0]
    assert(result['wasGeneratedBy'][id_dep]['prov:entity'] == 'dep:this')
    assert(result['wasGeneratedBy'][id_dep]['prov:activity'] == activity_id)

def assert_agent(result, agent_did):
    assert(result['agent'])
    assert(result['agent'][agent_did])
    assert(result['agent'][agent_did]['prov:type'])
    assert(result['agent'][agent_did]['prov:type']['$'] == 'dep:service-provider')

def assert_was_associated_with(result, activity_id, agent_did):
    assert(result['wasAssociatedWith'])
    id_dep = list(result['wasAssociatedWith'].keys())[0]
    assert(result['wasAssociatedWith'][id_dep]['prov:agent'] == agent_did)
    assert(result['wasAssociatedWith'][id_dep]['prov:activity'] == activity_id)

def assert_was_derived_from(result, asset_list):
    assert(result['wasDerivedFrom'])
    found_list = []
    for random_key, item in result['wasDerivedFrom'].items():
        assert(item['prov:generatedEntity'] == 'dep:this')
        assert(item['prov:usedEntity'] in asset_list)
        found_list.append(item['prov:usedEntity'])

    assert(len(found_list) == len(asset_list))

def test_provenance_create_publish():

    # minimum allowed
    provenance = Provenance()
    result = provenance.create_publish
    assert_prefix(result)
    activity_id = get_activity_id(result)
    assert_activity(result, activity_id, 'publish')
    assert_was_generated_by(result, activity_id)
    assert('agent' not in result)
    assert('wasAssociatedWith' not in result)

    # with agent_did
    agent_did = did_generate_random()
    provenance = Provenance(agent_did)

    result = provenance.create_publish
    assert_prefix(result)
    activity_id = get_activity_id(result)

    assert_activity(result, activity_id, 'publish')
    assert_was_generated_by(result, activity_id)

    assert_agent(result, agent_did)
    assert_was_associated_with(result, activity_id, agent_did)


    # with agent did and activity_id
    agent_did = did_generate_random()
    activity_id = 'abc-123-test'
    provenance = Provenance(agent_did=agent_did, activity_id=activity_id)
    result = provenance.create_publish
    #print(json.dumps(result, indent=4))
    assert_prefix(result)

    assert_activity(result, activity_id, 'publish')
    assert_was_generated_by(result, activity_id)

    assert_agent(result, agent_did)
    assert_was_associated_with(result, activity_id, agent_did)


def test_provenance_create_import():

    # minimum allowed
    provenance = Provenance()

    result = provenance.create_import
    assert_prefix(result)
    activity_id = get_activity_id(result)
    assert_activity(result, activity_id, 'import')
    assert_was_generated_by(result, activity_id)
    assert('agent' not in result)
    assert('wasAssociatedWith' not in result)

    # with agent_did
    agent_did = did_generate_random()

    provenance = Provenance(agent_did=agent_did)
    result = provenance.create_import
    assert_prefix(result)
    activity_id = get_activity_id(result)

    assert_activity(result, activity_id, 'import')
    assert_was_generated_by(result, activity_id)

    assert_agent(result, agent_did)
    assert_was_associated_with(result, activity_id, agent_did)


    # with agent did and activity_id
    agent_did = did_generate_random()
    activity_id = 'abc-123-test'
    provenance = Provenance(agent_did=agent_did, activity_id=activity_id)
    result = provenance.create_import
    assert_prefix(result)

    assert_activity(result, activity_id, 'import')
    assert_was_generated_by(result, activity_id)

    assert_agent(result, agent_did)
    assert_was_associated_with(result, activity_id, agent_did)


def test_provenance_create_invoke():
    # test minimum allowed for create_invoke
    provenance = Provenance()
    result = provenance.create_invoke
    assert_prefix(result)

    activity_id = get_activity_id(result)
    assert_activity(result, activity_id, 'invoke')
    assert_was_generated_by(result, activity_id)
    assert_entity(result)
    assert('agent' not in result)
    assert('wasAssociatedWith' not in result)


    #test all values set for create_invoke
    agent_did = did_generate_random()
    activity_id = 'job-id-123-abc'

    asset_list = []
    for index in range(0, 4):
        asset_id = secrets.token_hex(32)
        asset_list.append(f'{agent_did}/{asset_id}')

    inputs_text = json.dumps(
        {
            'asset_list': 'json'
        }
    )
    provenance = Provenance(agent_did=agent_did, activity_id=activity_id, asset_list=asset_list, inputs_text=inputs_text)
    result = provenance.create_invoke
    assert_prefix(result)

    activity_id = get_activity_id(result)
    assert_activity(result, activity_id, 'invoke', inputs_text)
    assert_was_generated_by(result, activity_id)
    assert_entity(result, asset_list)
    assert_agent(result, agent_did)
    assert_was_associated_with(result, activity_id, agent_did)
    assert_was_derived_from(result, asset_list)



"""

    Agent Module


"""

# from starfish.agent.agent_base import AgentBase                         # noqa: F401
from starfish.agent.agent_manager import AgentManager                   # noqa: F401
from starfish.agent.memory_agent import MemoryAgent                     # noqa: F401
from starfish.agent.remote_agent import RemoteAgent                     # noqa: F401



"""


Contract Base

"""

from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.convex.convex_network import ConvexNetwork


class ContractBase:
    def __init__(self, convex: ConvexNetwork, name: str, version: str):
        self._convex = convex
        self._name = name
        self._version = version
        self._source = None
        self._address = None

    def deploy(self, account: ConvexAccount):
        if not self._source:
            raise ValueError(f'Cannot deploy the contract {self.name} with no source text')

        deploy_line = f"""
(def {self.name}
    (deploy-once
        (quote
            (do
                {self._source}
            )
        )
    )
)"""
        result = self._convex.send(deploy_line, account)
        if result and 'value' in result:
            self._address = result['value']
            return self._address

    def load(self, deploy_account: str):
        self._address = self.get_address(deploy_account)
        self._version = self.get_version(deploy_account)
        return (self._address, self._version)

    def get_address(self, deploy_account):
        address = self._convex.get_address(self._name, deploy_account)
        return address

    def get_version(self, deploy_account):
        address = self._address
        if address is None:
            address = self.get_address(deploy_account)

        result = self._convex.query(f'(call {address} (version))', deploy_account)
        if result and 'value' in result:
            return result['value']

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def address(self):
        return self._address



"""
DID utils

"""

import re
import secrets
import warnings
from typing import Any

from urllib.parse import urlparse
from eth_utils import remove_0x_prefix

from web3 import Web3

from starfish.ddo import create_ddo_object
from starfish.types import DIDParts

NETWORK_DID_METHOD = 'dep'


def did_validate(did: str) -> bool:
    """

    Validate a did string.

    :param str did: DID string to validate, this can be an agent did or an asset_did
    :return: True if the did is valid
    :raise: TypeError if did is not a string
    :raise: ValueError if the DID is invalid

    """
    if not isinstance(did, str):
        raise TypeError('Expecting DID of string type, got %s of %s type' % (did, type(did)))

    if not re.match('^did:', did, re.IGNORECASE):
        raise ValueError(f'DID {did} must start with the text "did"')

    if not re.match('^did:([a-z0-9]+):', did, re.IGNORECASE):
        raise ValueError(f'DID {did} "id" must have only a-z 0-9 characters')

    if not re.match('^did:[a-z0-9]+:[a-f0-9]{64}.*', did, re.IGNORECASE):
        raise ValueError(f'DID {did} path should only have hex characters')
    return True


def is_did(did: str) -> bool:
    """

    Return True or False if the DID is valid.

    :param str did: DID string to validate, this can be an agent did or an asset_did

    """
    try:
        return did_validate(did)
    except (ValueError, TypeError):
        pass
    return False


def asset_did_validate(asse_did: str) -> bool:
    """

    Validates a asset_did in the format 'did:dep:<64_hex_chars>/<64_hex_chars>'

    :param str asse_did: Asset did to validate
    :return: True if validation is correct
    :raises: TypeError if did is not a string
    :raises: ValueError if the DID or the asset_id part of the DID is not valid

    """
    value = did_parse(asse_did)
    if value['path'] is None:
        raise ValueError(f'DID {asse_did} does not have an asset_id')
    if not re.match(r'^[0x]{0,2}[a-f0-9]{64}$', value['path']):
        raise ValueError(f'DID {asse_did} has an invalid asset_id')
    return True


def is_asset_did(asse_did: str) -> bool:
    """

    Return True or False if the asset_did is valid

    """
    try:
        return asset_did_validate(asse_did)
    except (ValueError, TypeError):
        pass
    return False


def did_parse(did: str) -> DIDParts:
    """

    parse a DID into it's parts.

    did:<method>:<id/id_hex>[/<path>[#<fragment>]]

    :param str did: DID to parse into seperate components
    :return: Dict of parts they are:
        method,
        id,
        id_hex,
        path,
        fragment


    """

    did_validate(did)

    match = re.match('^did:([a-z0-9]+):([a-f0-9]{64})(.*)', did, re.IGNORECASE)
    result = {
        'method': match.group(1),
        'id': match.group(2),
        'path': None,
        'fragment': None,
        'id_hex': None
    }

    uri_text = match.group(3)
    if uri_text:
        uri = urlparse(uri_text)
        result['fragment'] = uri.fragment
        if uri.path:
            result['path'] = uri.path[1:]

    if result['method'] == NETWORK_DID_METHOD and re.match('^[0-9a-f]{1,64}$', result['id'], re.IGNORECASE):
        result['id_hex'] = Web3.toHex(hexstr=result['id'])

    if not result['id_hex'] and result['id'].startswith('0x'):
        result['id_hex'] = result['id']

    return result


def did_generate_random() -> str:
    """

    Return a randomly generated DID

    """
    did_id = secrets.token_hex(32)
    return id_to_did(did_id)


def did_to_id(did: str) -> str:
    """

    Convert a DID to an hex id value

    """

    data = did_parse(did)
    return data['id_hex']


def id_to_did(did_id: str) -> str:
    """

    Convert an hex id to a DID

    """
    did_id = remove_0x_prefix(did_id)
    return f'did:{NETWORK_DID_METHOD}:{did_id}'


def decode_to_asset_id(asset_did_id: str) -> str:
    """

    Try to decode an id to an asset_id. The id can be a full asset_did, asset_id ( 32 hex number)

    :param str asset_did_id: string to try to decode
    :return: asset_id as a hex string, with no leading '0x'
    :raises ValueError: if no path exists in the asset_did or the path is invalid


    """

    asset_id = None
    if re.match(r'^[0-9a-fx]+$', asset_did_id, re.IGNORECASE):
        asset_id = Web3.toHex(hexstr=asset_did_id)
    else:
        data = did_parse(asset_did_id)
        if data['path'] is None:
            raise ValueError(f'Unable to get an asset_id from an agent DID address {asset_did_id}')

        if data['path'] and re.match('^[0-9a-fx]{1,66}$', data['path'], re.IGNORECASE):
            asset_id = data['path']
        else:
            raise ValueError(f'DID with asset_id is not valid {asset_did_id}')
    return remove_0x_prefix(asset_id)


def did_to_asset_id(did: str) -> str:
    """

    Depretaated function, use `decode_to_asset_id` instead
    """
    warnings.warn('use "decode_to_asset_id" instead', DeprecationWarning)
    return decode_to_asset_id(did)


def get_did_from_ddo(ddo_data: Any) -> str:
    ddo = create_ddo_object(ddo_data)
    if ddo:
        return ddo.did
    return None



"""

    Convex Contracts

"""

from starfish.network.convex.contract.contract_manager import ContractManager                               # noqa: F401
from starfish.network.convex.contract.ddo_registry_contract import DDORegistryContract                      # noqa: F401



"""

    Command Account Get ..

"""
from typing import Any
from web3 import Web3


from .command_base import CommandBase

DEFAULT_AMOUNT = 10


class AccountBalanceCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('balance', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Get balance of ether and tokens for an account address',
            help='Get balance of an account'

        )

        parser.add_argument(
            'address',
            help='Account address to get the balance'
        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        if not Web3.isAddress(args.address):
            output.add_error(f'{args.address} is not an ethereum account address')
            return

        network = self.get_network(args.url)
        ether_balance = network.get_ether_balance(args.address)
        token_balance = network.get_token_balance(args.address)
        output.add_line(f'ether balance: {ether_balance}')
        output.add_line(f'token balance: {token_balance}')
        output.set_value('ether', ether_balance)
        output.set_value('token', token_balance)



"""


    Contract Manager for Convex contracts


"""
import importlib
import inspect

from typing import Any

from starfish.network.convex.contract.contract_base import ContractBase
from starfish.network.convex.convex_network import ConvexNetwork
from starfish.types import TContractBase

CONTRACT_ACCOUNTS = {
    'development': '0x5288fec4153b702430771dfac8aed0b21cafca4344dae0d47b97f0bf532b3306',
}


class ContractManager:

    def __init__(self, convex: ConvexNetwork,  default_package_name: str):
        self._convex = convex
        self._default_package_name = default_package_name

    def load(self, name: str, package_name: str = None) -> TContractBase:
        """

        Load a contract using it's name, and network name

        :param str name: Name of the contract to load
        :param str artifact_filename: Optional filename of the artifact file
        :param bool has_artifact: Defaults to True, if false then just load the contract without an abi & address
        :param str package_name: Defaults to default_package_name, the package for the contract module

        :return: Return a contract object, else return None

        """
        if package_name is None:
            package_name = self._default_package_name
        if package_name is None:
            raise ValueError('You need to provide a package name for the contract classes')

        package_module = importlib.import_module(package_name)
        for item in inspect.getmembers(package_module, inspect.ismodule):
            class_def = ContractManager._find_class_in_module(name, item[1])
            if class_def:
                contract_object = class_def(self._convex)
                contract_object.load(CONTRACT_ACCOUNTS['development'])
                return contract_object
        return None

    @staticmethod
    def _find_class_in_module(class_name: str, contract_module: str) -> Any:
        for name, obj in inspect.getmembers(contract_module, inspect.isclass):
            if issubclass(obj, ContractBase) \
               and name != 'ContractBase' \
               and name == class_name:
                return obj
        return None



"""

Memory Agent class to provide basic functionality for Ocean Agents

"""

import json
import re
import secrets
from typing import Any

from starfish.agent.agent_base import AgentBase
from starfish.ddo.ddo import DDO
from starfish.listing import Listing
from starfish.purchase import Purchase
from starfish.types import (
    ListingData,
    TAsset,
    TListing
)
from starfish.utils.crypto_hash import hash_sha3_256
from starfish.utils.did import (
    decode_to_asset_id,
    did_generate_random
)


class MemoryAgent(AgentBase):
    """

    Memory Agent class allows to register, list, purchase and consume assets.

    :param ddo: ddo to access the agent
    :type ddo: :class:`.DDO`

    """

    def __init__(self, ddo: DDO = None) -> None:

        if ddo is None:
            did = did_generate_random()
            ddo = DDO(did)

        AgentBase.__init__(self, ddo=ddo)

        self._memory = {
            'listing': {},
            'asset': {},
            'purchase': {}
        }

    def register_asset(self, asset: TAsset) -> TAsset:
        """

        Register a memory asset.

        :param object metadata: asset object to register for this asset.

        :return: A :class:`.AssetBase` object that has been registered, if failure then return None.
        :type: :class:`.AssetBase` class

        """

        asset_id = hash_sha3_256(asset.metadata_text)
        did = f'{self._ddo.did}/{asset_id}'

        self._memory['asset'][did] = asset
        asset.set_did(did)
        return asset

    def create_listing(self, listing_data: ListingData, asset_did: str) -> TListing:
        """

        Create a listing on the market place for this asset

        :param dict listing_data:  Listing inforamiton to give for this asset
        :param str asset_did: asset DID to assign to this listing
        :return: A new :class:`.Listing` object that has been registered, if failure then return None.
        :type: :class:`.Listing` class

        """
        listing = None
        if listing_data:
            listing_id = decode_to_asset_id(asset_did)
            listing_data['listing_id'] = listing_id,
            listing_data['asset_did'] = asset_did
            listing = Listing(self, listing_id, asset_did, listing_data)
            self._memory['listing'][listing_id] = listing
        return listing

    def update_listing(self, listing: TListing) -> None:
        """

        Update the listing to the agent server.

        :param listing: Listing object to update
        :type listing: :class:`.Listing` class

        """
        self._memory['listing'][listing.listing_id] = listing

    def validate_asset(self, asset: TAsset) -> bool:
        """

        Validate an asset

        :param asset: Asset to validate.
        :return: True if the asset is valid
        """
        return asset is not None

    def get_listing(self, listing_id: str) -> TListing:
        """

        Return an listing from the given listing_id.

        :param str listing_id: Id of the listing.

        :return: a registered listing given a Id of the listing
        :type: :class:`.Listing` class

        """
        listing = None
        if listing_id in self._memory['listing']:
            listing = self._memory['listing'][listing_id]
        return listing

    def search_listings(self, text: str, sort: str = None, offset: int = 100, page: int = 0) -> Any:
        """

        Search for listings with the givien 'text'

        :param str text: Text to search all listing data for.
        :param sort: sort the results ( defaults: None, no sort).
        :type sort: str or None
        :param int offset: Return the result from with the maximum record count ( defaults: 100 ).
        :param int page: Returns the page number based on the offset.

        :return: a list of listing objects found using the search.
        :type: :class:`.Listing` objects

        For example::

            # return the 300 -> 399 records in the search for the text 'weather' in the metadata.
            my_result = agent.search_registered_assets('weather', None, 100, 3)

        """
        listing_id_list = []
        for listing_id, listing in self._memory['listing'].items():
            if re.search(text, json.dumps(listing.data)):
                listing_id_list.append(listing.listing_id)

        return listing_id_list

    def purchase_asset(self, listing: Any, account: Any) -> Any:
        """

        Purchase an asset using it's listing and an account.

        :param listing: Listing to use for the purchase.
        :type listing: :class:`.Listing`
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        """
        purchase = None

        purchase_id = secrets.token_hex(64)
        if purchase_id:
            purchase = Purchase(self, listing, purchase_id, account)
            self._memory['purchase'][purchase_id] = (purchase, account.address)

        return purchase

    def purchase_wait_for_completion(self, asset: Any, account: Any,  purchase_id: str, timeoutSeconds: int) -> bool:
        """

            Wait for completion of the purchase

            TODO: issues here...
            + No method as yet to pass back paramaters and values during the purchase process
            + We assume that the following templates below will always be used.

        """
        return True

    def is_access_granted_for_asset(self, asset: Any, account: Any, purchase_id: str = None) -> bool:
        """

        Check to see if the account and purchase_id have access to the assed data.


        :param asset: Asset to check for access.
        :type asset: :class:`.Asset` object
        :param account: Ocean account to purchase the asset.
        :type account: :class:`.Account` object to use for registration.
        :param str purchase_id: purchase id that was used to purchase the asset.

        :return: True if the asset can be accessed and consumed.
        :type: boolean
        """

        if purchase_id in self._memory['purchase']:
            purchase, account_address = self._memory['purchase'][purchase_id]
            return purchase and account.is_address_equal(account_address)

        return False

    def get_asset_purchase_ids(self, asset: Any) -> Any:
        """

        Returns as list of purchase id's that have been used for this asset

        :param asset: DataAsset to return purchase details.
        :type asset: :class:`.DataAsset` object

        :return: list of purchase ids
        :type: list

        """
        return []

    def consume_asset(self, listing: Any, account: Any, purchase_id: str) -> bool:
        """
        Consume the asset and download the data. The actual payment to the asset
        provider will be made at this point.

        :param listing: Listing that was used to make the purchase.
        :type listing: :class:`.Listing`
        :param str purchase_id: purchase id that was used to purchase the asset.
        :param account: Ocean account that was used to purchase the asset.
        :type account: :class:`.Account` object to use for registration.

        :return: True if the asset has been consumed and downloaded
        :type: boolean

        """
        return purchase_id in self._memory['purchase']



"""

    Command 'Agent Register'

"""
import re
from typing import Any

from starfish.agent import RemoteAgent
from starfish.agent.services import (
    ALL_SERVICES,
    Services
)
from starfish.network.ethereum.ethereum_account import EthereumAccount


from .command_base import CommandBase


class AgentRegisterCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('register', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Tool to register an agent',
            help='Tool to register an agent on the network',

        )

        parser.add_argument(
            'agent_url',
            help='url of the agent to register, you may need to provide --username and --password options'
        )

        parser.add_argument(
            'password',
            help='account password to use to register'
        )

        parser.add_argument(
            'keyfile',
            help='account key_file use to register'
        )

        parser.add_argument(
            'service_list',
            nargs='?',
            help=f'''
Optional list of services, if not given then all services will be registed.
Services can be: {",".join(ALL_SERVICES)}
'''
        )
        return parser

    def execute(self, args: Any, output: Any) -> Any:
        network = self.get_network(args.url)

        register_account = EthereumAccount.import_from_file(args.keyfile, args.password)

        all_services = True
        service_list = None
        if args.service_list:
            name_list = re.split(r'[^a-z]+', args.service_list, re.IGNORECASE)
            service_list = []
            print(name_list)
            for name in name_list:
                service_name = name.strip()
                if service_name not in ALL_SERVICES:
                    output.add_error(f'{service_name} is not a valid service')
                    return
                service_list.append(service_name)
            if service_list:
                all_services = False
            else:
                output.add_error(f'No services found to register in list {args.service_list}')
                return

        services = Services(args.agent_url, service_list=service_list, all_services=all_services)
        ddo = RemoteAgent.generate_ddo(services)
        if network.register_did(register_account, ddo.did, ddo.as_text()):
            output.add_line(f'{args.agent_url} has been registered with the did {ddo.did}')
            output.add_line(ddo.as_text())
            output.set_value('did', ddo.did)
            output.set_value('ddo_text', ddo.as_text())
        else:
            output.add_error(f'unable to register {args.agent_url}')



"""
DEX DID/DDO Python Library

.. moduleauthor: Bill


This version was copied from

https://github.com/oceanprotocol/squid-py/tree/4dc75b959f9f07101ff6ff30d33b4d95b9672834/squid_py/ddo


"""
from typing import Any

from .ddo import DDO                     # noqa: F401

from .public_key_base import (               # noqa: F401
    PUBLIC_KEY_STORE_TYPE_BASE64,
    PUBLIC_KEY_STORE_TYPE_BASE85,
    PUBLIC_KEY_STORE_TYPE_HEX,
    PUBLIC_KEY_STORE_TYPE_PEM
)


def create_ddo_object(ddo_data: Any) -> Any:
    ddo = None
    if isinstance(ddo_data, str):
        ddo = DDO(json_text=ddo_data)
    elif isinstance(ddo_data, dict):
        ddo = DDO(dictionary=ddo_data)
    elif isinstance(ddo_data, DDO):
        ddo = ddo_data
    return ddo



"""
    Public key base

    Currently this class stores the public keys in the same form as in the JSON
    text data.

"""

import json
import re
from base64 import (
    b64decode,
    b64encode,
    b85decode,
    b85encode
)
from typing import Any

PUBLIC_KEY_STORE_TYPE_PEM = 'publicKeyPem'
PUBLIC_KEY_STORE_TYPE_JWK = 'publicKeyJwk'
PUBLIC_KEY_STORE_TYPE_HEX = 'publicKeyHex'
PUBLIC_KEY_STORE_TYPE_BASE64 = 'publicKeyBase64'
PUBLIC_KEY_STORE_TYPE_BASE85 = 'publicKeyBase85'


class PublicKeyBase:
    """
    Base Public Key class, to allow to perfom basic key storage and validation
    using DDO keys
    """

    def __init__(self, key_id: str, **kwargs: Any) -> None:
        self._id = key_id
        self._store_type = kwargs.get('store_type', None)
        self._value = kwargs.get('value', None)
        self._owner = kwargs.get('owner', None)
        self._type = kwargs.get('type', None)

    @property
    def id(self) -> str:
        """ get the key id"""
        return self._id

    def assign_did(self, did: str) -> None:
        """
        assign the DID as the key id, if the DID does not have a '#value'
        at the end, then automatically add a new key value
        """
        if re.match('^#.*', self._id):
            self._id = did + self._id
        if re.match('^#.*', self._owner):
            self._owner = did + self._owner

    @property
    def owner(self) -> str:
        """get the owner of this key"""
        return self._owner

    @property
    def type(self) -> str:
        """get the type of key"""
        return self._type

    @property
    def store_type(self) -> str:
        """get the type of key storage"""
        return self._store_type

    def set_key_value(self, value: str, store_type: str = PUBLIC_KEY_STORE_TYPE_BASE64) -> None:
        """set the key value based on it's storage type"""
        if isinstance(value, dict):
            if PUBLIC_KEY_STORE_TYPE_HEX in value:
                self.set_key_value(value[PUBLIC_KEY_STORE_TYPE_HEX], PUBLIC_KEY_STORE_TYPE_HEX)
            elif PUBLIC_KEY_STORE_TYPE_BASE64 in value:
                self.set_key_value(value[PUBLIC_KEY_STORE_TYPE_BASE64], PUBLIC_KEY_STORE_TYPE_BASE64)
            elif PUBLIC_KEY_STORE_TYPE_BASE85 in value:
                self.set_key_value(value[PUBLIC_KEY_STORE_TYPE_BASE85], PUBLIC_KEY_STORE_TYPE_BASE85)
            elif PUBLIC_KEY_STORE_TYPE_JWK in value:
                self.set_key_value(value[PUBLIC_KEY_STORE_TYPE_JWK], PUBLIC_KEY_STORE_TYPE_JWK)
            elif PUBLIC_KEY_STORE_TYPE_PEM in value:
                self.set_key_value(value[PUBLIC_KEY_STORE_TYPE_PEM], PUBLIC_KEY_STORE_TYPE_PEM)
        else:
            self._value = value
            self._store_type = store_type

    def set_encode_key_value(self, value: str, store_type: str) -> str:
        """ save the key value base on it's storage type"""
        self._store_type = store_type
        if store_type == PUBLIC_KEY_STORE_TYPE_HEX:
            self._value = value.hex()
        elif store_type == PUBLIC_KEY_STORE_TYPE_BASE64:
            self._value = b64encode(value).decode()
        elif store_type == PUBLIC_KEY_STORE_TYPE_BASE85:
            self._value = b85encode(value).decode()
        elif store_type == PUBLIC_KEY_STORE_TYPE_JWK:
            # TODO: need to decide on which jwk library to import?
            raise NotImplementedError
        else:
            self._value = value
        return value

    @property
    def decode_value(self) -> str:
        """ return the key value based on it's storage type"""
        if self._store_type == PUBLIC_KEY_STORE_TYPE_HEX:
            value = bytes.fromhex(self._value)
        elif self._store_type == PUBLIC_KEY_STORE_TYPE_BASE64:
            value = b64decode(self._value)
        elif self._store_type == PUBLIC_KEY_STORE_TYPE_BASE85:
            value = b85decode(self._value)
        elif self._store_type == PUBLIC_KEY_STORE_TYPE_JWK:
            # TODO: need to decide on which jwk library to import?
            raise NotImplementedError
        else:
            value = self._value
        return value

    @property
    def value(self) -> str:
        """ get the key value"""
        return self._value

    def as_text(self, is_pretty: bool = False) -> str:
        """ return the key as JSON text"""
        values = {'id': self._id, 'type': self._type, self._store_type: self._value}
        if self._owner:
            values['owner'] = self._owner

        if is_pretty:
            return json.dumps(values, indent=4, separators=(',', ': '))

        return json.dumps(values)

    @property
    def as_dictionary(self) -> Any:
        """return the key as a python dictionary"""
        values = {
            'id': self._id,
            'type': self._type,
        }
        if self._store_type:
            values[self._store_type] = self._value

        if self._owner:
            values['owner'] = self._owner
        return values

    @property
    def is_valid(self) -> bool:
        """return True if the key structure is valid"""
        return self._id and self._type

    @property
    def authentication_type(self) -> str:
        """
        base overloaded method to return the authentication type to use for
        this key
        """
        raise NotImplementedError



"""

    Command Account Get Ether

"""
from typing import Any
from web3 import Web3

from starfish.network.ethereum.ethereum_account import EthereumAccount

from .command_base import CommandBase


DEFAULT_AMOUNT = 10


class AccountGetEtherCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        super().__init__('ether', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:

        parser = sub_parser.add_parser(
            self._name,
            description='Get ether for an account address',
            help='Get some ether from a faucet or transfered on a development network'

        )

        parser.add_argument(
            'address',
            help='Account address'
        )

        parser.add_argument(
            'faucet',
            metavar='network-faucet-url',
            help='Network name or faucet url'
        )

        parser.add_argument(
            'amount',
            nargs='?',
            default=DEFAULT_AMOUNT,
            help=f'Amount to request. Default {DEFAULT_AMOUNT}'
        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        if not Web3.isAddress(args.address):
            output.add_error(f'{args.address} is not an ethereum account address')
            return

        account = EthereumAccount(args.address)

        faucet_url = args.faucet
        faucet_account = None
        network_setup = self.get_network_setup(args.faucet)
        if network_setup:
            if 'faucet_url' in network_setup:
                faucet_url = network_setup['faucet_url']
            if 'faucet_account' in network_setup:
                faucet_account = network_setup['faucet_account']
            network = self.get_network(args.url, network_setup['url'])
        else:
            network = self.get_network(args.url)

        if faucet_account:
            send_account = EthereumAccount(faucet_account[0], faucet_account[1])
            network.send_ether(send_account, account.address, args.amount)
        else:
            network.request_ether_from_faucet(account, faucet_url)

        balance = network.get_ether_balance(account)
        output.add_line(f'Get {args.get_command} for account: {args.address} balance: {balance}')
        output.set_value('balance', balance)
        output.set_value('address', args.address)



"""

    Network class to provide basic functionality for convex network

"""
from convex_api import ConvexAPI

from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.network_base import NetworkBase
from starfish.types import AccountAddress

DEFAULT_PACKAGE_NAME = 'starfish.network.convex.contract'


class ConvexNetwork(NetworkBase):
    def __init__(self, url: str) -> None:
        NetworkBase.__init__(self, url)
        self._convex = ConvexAPI(url)
        from starfish.network.convex.contract.contract_manager import ContractManager
        self._manager = ContractManager(self._convex, DEFAULT_PACKAGE_NAME)

    """

    Register DID with a DDO and resolve DID to a DDO

    """
    def register_did(self, did: str, ddo_text: str, account: ConvexAccount) -> bool:
        ddo_registry_contract = self._manager.load('DDORegistryContract')
        if ddo_registry_contract:
            return ddo_registry_contract.register_did(did, ddo_text, account)

    def resolve_did(self, did: str, account_address: AccountAddress) -> str:
        ddo_registry_contract = self._manager.load('DDORegistryContract')
        if ddo_registry_contract:
            return ddo_registry_contract.resolve(did,  account_address)
        return None

    @property
    def convex(self):
        return self._convex



"""
    starfish-ddo-registry contract

"""

from starfish.network.convex.contract.contract_base import ContractBase
from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.convex.convex_network import ConvexNetwork
from starfish.types import AccountAddress


class ProvenanceContract(ContractBase):

    def __init__(self, convex: ConvexNetwork):
        ContractBase.__init__(self, convex, 'starfish-provenance', '0.0.1')

        self._source = f'''
            (def provenance [])
            (defn version [] "{self.version}")
            (defn assert-asset-id [value]
                (when-not (and (blob? value) (== 32 (count (blob value)))) (fail "INVALID" "invalid asset-id"))
            )
            (defn assert-address [value]
                (when-not (address? (address value)) (fail "INVALID" "invalid address"))
            )
            (defn register [asset-id]
                (assert-asset-id asset-id)
                (let [record {{:owner *caller* :timestamp *timestamp* :asset-id (blob asset-id)}}]
                    (def provenance (conj provenance record))
                    record
                )
            )
            (defn event-list [asset-id]
                (assert-asset-id asset-id)
                (mapcat (fn [record] (when (= (blob asset-id) (record :asset-id)) [record])) provenance)
            )
            (defn event-owner [owner-id]
                (assert-address owner-id)
                (mapcat (fn [record] (when (= (address owner-id) (record :owner)) [record])) provenance)
            )
            (defn event-timestamp [time-from time-to]
                (mapcat
                    (fn [record]
                        (when
                            (and
                                (<= time-from (record :timestamp))
                                (>= time-to (record :timestamp))
                            )
                        [record] )
                    )
                provenance)
            )
            (export event-list event-owner event-timestamp register version)

'''

    def register_did(self, asset_id: str, account: ConvexAccount):
        command = f'(call {self.address} (register {asset_id}))'
        result = self._convex.send(command, account)
        if result and 'value' in result:
            return result['value']
        return result

    def event_list(self, asset_id: str, account_address: AccountAddress):
        command = f'(call {self.address} (event_list {asset_id}))'
        if isinstance(account_address, str):
            address = account_address
        else:
            address = account_address.address
        result = self._convex.query(command, address)
        if result and 'value' in result:
            return result['value']
        return result



"""

    Command 'get_ether'

"""
from typing import Any

from .account_balance_command import AccountBalanceCommand
from .account_create_command import AccountCreateCommand
from .account_get_command import AccountGetCommand
from .account_send_command import AccountSendCommand
from .command_base import CommandBase
from .help_command import HelpCommand


class AccountCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        self._command_list = []
        super().__init__('account', sub_parser)

    def create_parser(self, sub_parser: Any) -> Any:
        parser = sub_parser.add_parser(
            self._name,
            description='Tool tasks on accounts',
            help='Tasks to perform on accounts',

        )
        account_parser = parser.add_subparsers(
            title='Account sub command',
            description='Account sub command',
            help='Account sub command',
            dest='account_command'
        )

        self._command_list = [
            AccountBalanceCommand(account_parser),
            AccountCreateCommand(account_parser),
            AccountGetCommand(account_parser),
            AccountSendCommand(account_parser),
            HelpCommand(account_parser, self)
        ]
        return account_parser

    def execute(self, args: Any, output: Any) -> None:
        return self.process_sub_command(args, output, args.account_command)



"""

    Agent Manager, to load and resolve remote agents

"""
import logging
from typing import Dict

from starfish.agent.remote_agent import RemoteAgent
from starfish.ddo import create_ddo_object
from starfish.ddo.ddo import DDO
from starfish.network.network_base import NetworkBase
from starfish.types import (
    AgentItem,
    Authentication,
    TRemoteAgent
)
from starfish.utils.did import (
    did_to_id,
    id_to_did
)

logger = logging.getLogger(__name__)


class AgentManager:
    def __init__(self, network: NetworkBase) -> None:
        self._network = network
        self._items = {}
        self._default_name = None

    def add(
        self,
        name: [str, dict],
        url: str = None,
        did: str = None,
        authentication: Authentication = None,
        is_default: bool = None
    ) -> None:
        """

        Add a remote agent details to the list of remote agents managed by this class.

        :param str name: name of the remote agent
        :param str,dict url: url of the remote agent, or a dict if passing a dict of items
        :param str did: DID of the agent
        :param dict authentication: Access details to pass to the agent

        """
        if url and isinstance(url, dict):
            item = url
            if 'username' in item or 'token' in item:
                authentication = {
                    'username': item.get('username', None),
                    'password': item.get('password', None),
                    'token': item.get('token', None)
                }
                item['authentication'] = authentication
        else:
            item = {
                'url': url,
                'did': did,
                'authentication': authentication,
            }
        if item['url'] is None and item['did'] is None:
            raise ValueError('You must provide at least an URL or DID for the remote agent')

        # if the is_default value is set to True, then make this the default agent
        if is_default:
            self._default_name = is_default

        # if _default_name is not assigned, then assign it to the first name
        if self._default_name is None:
            self._default_name = name

        self._items[name] = item

    def resolve_ddo(self, name: str) -> str:
        """

        Tries to resolve the remote agent to a ddo using it's provided URL or DID.

        :param str name: Name of the remote agent that was added to the managed list.
        :return dict: DDO text or None for the remote agent

        """

        if name not in self._items:
            raise ValueError(f'remote agent {name} not found in list')

        item = self._items[name]
        if item.get('ddo_text', None) is None:

            ddo_text = self._network.resolve_did(item.get('did', None))
            if ddo_text is None:
                authentication = item.get('authentication', None)
                ddo_text = RemoteAgent.resolve_url(item.get('url', None), authentication=authentication)

            self._items[name]['ddo_text'] = ddo_text
            self._items[name]['ddo'] = create_ddo_object(ddo_text)
        return self._items[name]['ddo_text']

    def load_agent(self, asset_agent_did_name: str) -> TRemoteAgent:
        """

        Resolves and gets a valid remote agent for a given asset_did, agent_did or agent name

        :param str asset_agent_did_name:  Can be an asset_did, agent_did or name of the agent held by this object

        :return RemoteAgent: Object or None if none found

        """

        agent = None
        ddo_text = None

        # test for name in the agent list
        if asset_agent_did_name in self._items:
            logger.debug(f'found {asset_agent_did_name} in the list of agents')
            ddo_text = self.resolve_ddo(asset_agent_did_name)
        else:
            # find in list of did's
            # strip away the asset part of the id
            did_id = did_to_id(asset_agent_did_name)
            did = id_to_did(did_id)
            logger.debug(f'stripped possible did {asset_agent_did_name} to {did}')
            item = self.get_item_from_did(did)
            if item:
                logger.debug(f'found {did} in list of resolved items')
                ddo_text = item['ddo_text']
            else:
                # if it's a agent_did or asset_did
                ddo_text = self._network.resolve_did(did)
                if ddo_text:
                    logger.debug(f'resolved {did} from network')

        if ddo_text:
            ddo = DDO(json_text=ddo_text)
            authentication = None

            item = self.get_item_from_did(ddo.did)
            if item:
                authentication = item.get('authentication', None)
            agent = RemoteAgent(ddo_text, authentication=authentication)
        return agent

    def get_item_from_did(self, find_did: str) -> AgentItem:
        for name, item in self._items.items():
            did = None
            if 'ddo' in item:
                did = item['ddo'].did
            elif 'ddo_text' in item:
                ddo = create_ddo_object(item['ddo_text'])
                did = ddo.did
            if did and did == find_did:
                return item
        return None

    @property
    def default_name(self) -> str:
        return self._default_name

    @default_name.setter
    def default_name(self, value: str) -> None:
        self._default_name = value

    @property
    def items(self) -> Dict[str, AgentItem]:
        return self._items



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
from os.path import join

from setuptools import (
    setup,
    find_packages
)

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('CHANGELOG.md') as changelog_file:
    changelog = changelog_file.read()


install_requirements = [
    'web3',
    'typing_extensions',
    'mongoquery',
    'convex_api_py',
]

setup_requirements = ['pytest-runner', ]

test_requirements = [
    'flake8',
    'pytest',
    'isort',
    'mypy',
    'pyyaml',
]

# Possibly required by developers of starfish-py:
dev_requirements = [
    'bumpversion',
    'isort',
    'mypy',
]

docs_requirements = [
    'Sphinx',
    'sphinx-rtd-theme',
    'sphinxcontrib-apidoc',
    'sphinxcontrib-plantuml',
    'sphinx-automodapi',
    'pygments',
]

setup(
    author="dex-company",
    author_email='devops@dex.sg',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
    ],
    description="Developer Toolkit for Decentralised Data Ecosystems",
    extras_require={
        'test': test_requirements,
        'docs': docs_requirements,
        'dev': dev_requirements + test_requirements + docs_requirements,
    },
    install_requires=install_requirements,
    #dependency_links=dependency_links,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords='starfish-py',
    name='starfish-py',
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'starfish.contract': ['data/*.json.gz', 'data/*.json']
    },
    scripts=[
        'tools/starfish_tools',
    ],
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/DEX-Company/starfish-py',
    version='0.13.0',
    zip_safe=False,
)



"""
    starfish-ddo-registry contract

"""

from starfish.network.convex.contract.contract_base import ContractBase
from starfish.network.convex.convex_account import ConvexAccount
from starfish.network.convex.convex_network import ConvexNetwork
from starfish.types import AccountAddress


class DDORegistryContract(ContractBase):

    def __init__(self, convex: ConvexNetwork):
        ContractBase.__init__(self, convex, 'starfish-ddo-registry', '0.0.4')

        self._source = f'''
            (def registry {{}})
            (def creator *caller*)
            (defn version [] "{self.version}")
            (defn get-register [did] (get registry did) )
            (defn set-register [did owner-address ddo]
                (let [register-record {{:owner owner-address :ddo ddo}}]
                    (def registry (assoc registry did register-record))
                )
            )
            (defn delete-register [did] (def registry (dissoc registry did)) )
            (defn assert-owner [did]
                (when-not (owner? did) (fail "NOT-OWNER" "not owner"))
            )
            (defn assert-address [value]
                (when-not (address? (address value)) (fail "INVALID" "invalid address"))
            )
            (defn assert-did [value]
                (when-not (and (blob? value) (== 32 (count (blob value)))) (fail "INVALID" "invalid DID"))
            )
            (defn resolve? [did]
                (assert-did did)
                (boolean (get-register did))
            )
            (defn resolve [did]
                (assert-did did)
                (when-let [register-record (get-register did)] (register-record :ddo))
            )
            (defn owner [did]
                (assert-did did)
                (when-let [register-record (get-register did)] (register-record :owner))
            )
            (defn owner? [did] (= (owner did) *caller*) )
            (defn register [did ddo]
                (assert-did did)
                (when (resolve? did) (assert-owner did))
                (set-register did *caller* ddo)
                did
            )
            (defn unregister [did]
                (when (resolve? did)
                    (assert-owner did)
                    (delete-register did)
                    did
                )
            )
            (defn transfer [did to-account]
                (when (resolve? did)
                    (assert-owner did)
                    (assert-address to-account)
                    (set-register did (address to-account) (resolve did))
                    [did (address to-account)]
                )
            )
            (defn owner-list [the-owner]
                (assert-address the-owner)
                (mapcat (fn [v] (when (= (address the-owner) (get (last v) :owner)) [(first v)])) registry)
            )
            (export resolve resolve? register unregister owner owner? owner-list transfer version)
        '''

    def register_did(self, did: str, ddo_text: str, account: ConvexAccount):
        command = f'(call {self.address} (register {did} "{ddo_text}"))'
        result = self._convex.send(command, account)
        if result and 'value' in result:
            return result['value']
        return result

    def resolve(self, did: str, account_address: AccountAddress):
        command = f'(call {self.address} (resolve {did}))'
        if isinstance(account_address, str):
            address = account_address
        else:
            address = account_address.address
        result = self._convex.query(command, address)
        if result and 'value' in result:
            return result['value']
        return result

    def owner(self, did: str,  account_address: AccountAddress):
        command = f'(call {self.address} (owner {did}))'
        if isinstance(account_address, str):
            address = account_address
        else:
            address = account_address.address
        result = self._convex.query(command, address)
        if result and 'value' in result:
            return result['value']
        return result




from unittest.mock import Mock
import secrets

from tests.unit.libs.unit_test_account import UnitTestAccount

class ConfigEthereum:
    network_url = 'http://test-keeper-url:1234'
    accounts = []


class UnitTestConfig():

    remote_agent_url = 'http://localhost:8080'
    def __init__(self):
        self.ethereum = ConfigEthereum()
        for index in range(0, 2):
            account = self.create_account(secrets.token_hex(32))
            self.ethereum.accounts.append(account)

    def create_account(self, password):
        account = UnitTestAccount(password)
        return account

unitTestConfig = UnitTestConfig()



"""

    Tool command Account Create


"""
import logging
from typing import Any

from starfish.network.ethereum.ethereum_account import EthereumAccount
from .command_base import CommandBase

logger = logging.getLogger(__name__)


class AccountCreateCommand(CommandBase):

    def __init__(self, sub_parser: Any = None) -> None:
        super().__init__('create', sub_parser)

    def create_parser(self, sub_parser):
        parser = sub_parser.add_parser(
            self._name,
            description='Create a new account',
            help='Create a new account'

        )
        parser.add_argument(
            'password',
            help='Password of new account'
        )

        parser.add_argument(
            'keyfile',
            nargs='?',
            help='Optional name of the keyfile to write the account key data to'
        )
        return parser

    def execute(self, args: Any, output: Any) -> None:
        account = EthereumAccount.create_new(args.password)
        logger.debug(f'create new account {account.address}')
        if args.keyfile:
            logger.debug(f'writing key file to {args.keyfile}')
            account.save_to_file(args.keyfile)
        else:
            logger.debug('writing key file to ouptut')
            output.add_line(account.export_to_text)
        output.add_line(account.address)
        output.set_value('keydata', account.key_data)
        output.set_value('address', account.address)
